#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cmath>

// Macros for input checking
#define CHECK_CUDA(x) TORCH_CHECK(x.is_cuda(), #x " must be a CUDA tensor")
#define CHECK_CONTIGUOUS(x) TORCH_CHECK(x.is_contiguous(), #x " must be contiguous")
#define CHECK_INPUT(x) \
    CHECK_CUDA(x);     \
    CHECK_CONTIGUOUS(x)

// This kernel uses 2D grid and block indexing to directly map the 2D output
// of the linear transformation. Grid-stride loops in both row and column dimensions
// ensure that all elements are processed efficiently, even for larger tensors.
template <typename scalar_t>
__global__ void twoD_activation_kernel(
    const scalar_t* __restrict__ x_in,
    scalar_t* __restrict__ x_out,
    const int m,
    const int n) {

    // Calculate the starting row and column for this thread
    int col = blockIdx.x * blockDim.x + threadIdx.x;
    int row = blockIdx.y * blockDim.y + threadIdx.y;

    const scalar_t one = static_cast<scalar_t>(1);
    const scalar_t half = static_cast<scalar_t>(0.5);

    // Grid-stride loop in row and column directions
    for (int r = row; r < m; r += blockDim.y * gridDim.y) {
        for (int c = col; c < n; c += blockDim.x * gridDim.x) {
            int idx = r * n + c;
            scalar_t x = x_in[idx];
            
            // Swish activation: x * sigmoid(x)
            scalar_t sigmoid = one / (one + exp(-x));
            x = x * sigmoid;
            
            // Divide by 2
            x = x * half;
            
            // First clamp between -1 and 1
            x = max(min(x, one), -one);
            
            // Tanh activation
            x = tanh(x);
            
            // Second clamp between -1 and 1
            x = max(min(x, one), -one);
            
            x_out[idx] = x;
        }
    }
}

// CUDA forward function that fuses the linear transformation (using torch::addmm)
// with the activation operations using 2D grid indexing.
torch::Tensor module_forward_cuda(
    torch::Tensor x,
    torch::Tensor weight,
    torch::Tensor bias) {

    // Perform linear operation: x_linear = bias + x * weight^T
    auto x_linear = torch::addmm(bias, x, weight.t());
    auto x_out = torch::empty_like(x_linear);

    int m = x_linear.size(0);
    int n = x_linear.size(1);

    // Define 2D block and grid dimensions to match the tensor's 2D structure
    const dim3 block(32, 8);  // 256 threads per block
    const dim3 grid((n + block.x - 1) / block.x, (m + block.y - 1) / block.y);

    AT_DISPATCH_FLOATING_TYPES_AND_HALF(x_linear.scalar_type(), "module_forward_cuda", ([&] {
        twoD_activation_kernel<scalar_t><<<grid, block>>>(
            x_linear.data_ptr<scalar_t>(),
            x_out.data_ptr<scalar_t>(),
            m,
            n);
    }));

    return x_out;
}

// C++ interface
torch::Tensor module_forward(
    torch::Tensor x,
    torch::Tensor weight,
    torch::Tensor bias) {
    CHECK_INPUT(x);
    CHECK_INPUT(weight);
    CHECK_INPUT(bias);
    return module_forward_cuda(x, weight, bias);
}

// PyBind11 module definition
PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_forward, "2D grid based fused activation forward function (CUDA)");
}
